/***********************************************************************
 * File: csv_converter.h
 * 	(Extracted from
 *	   "$Id: csv_converter.h,v 1.1 2008/08/09 16:55:45 maoyam Exp maoyam $")
 *
 * Abstracts:
 *  This C source code implements the converter set for the CSV-files.
 *
 * Copyright (C) 2004-2007 Maoyam Tokyo, Japan (mailto:maoyam@mail.goo.ne.jp)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 **********************************************************************/

#if !defined(_CSV_CONVERTER_H_) /* { */
#define _CSV_CONVERTER_H_
#ifdef __cplusplus
extern "C" {
#endif

/* include the system definitions */
#include <unistd.h>	/* for definition NULL */
#include <stdio.h>	/* for declaration of `stderr' */
#include <errno.h>	/* for declaration of `errno' */
#include <stdlib.h>	/* for malloc(3C), realloc(3C) and free(3C)*/
#include <string.h>	/* for C-string objects */
#include <stdarg.h>	/* for variable arguments */

/* definitions for strings_to_csv_line() */
#define DLMT_COMMA ','	/* for Comma Separated Values (CSV) format */
#define DLMT_TAB   '\t'	/* for Tab Separated Values (TSV) format */
#define DLMT_FORBIDDEN_CHARS "\r\"\n"
			/* forbidden characters for using delimiter */

/* definitions for get_csv_line() */
#if defined(DEBUG) /* { */
#	define DEFAULT_CSV_LINE_SZ	(10)
#else  /* } !defined(DEBUG) { */
#	define DEFAULT_CSV_LINE_SZ	(4 * 1024)
#endif/* } !define(DEBUG) */

/* definitions for parse_csv_file() */
typedef struct parsed_csv_line_t {
	size_t	linebuf_sz;
			/* size of allocated memories for `linebuf' */
	char	*linebuf; /* pointing to allocated buffer */

	int	cols_cnt; /* count of the elements for `cols' vector */
	char	**cols;	  /* pointing to the vector of C-strings */
} Parsed_CSV_Line_t;	/* CSV-line */

typedef struct parsed_csv_t {
	int			line_cnt;
				/* count of elements for `lines' vector
				 */
	Parsed_CSV_Line_t	**lines;
				/* pointing to the vector of parsed
				 * CSV-lines
				 */

	int			min_cols_cnt;
				/* minimum count of the lines */
	int			max_cols_cnt;
				/* maximum count of the lines */
} Parsed_CSV_t;
		/* parsed structure for the CSV formatted file */

typedef enum errcode_parse_csv_file_t {
	  EPCF_NORMAL_END = 0
	, EPCF_Null_Pointer_Para	/* Illegal Parameters */
	, EPCS_Illegal_Delimiter	/* Unknown delimiter string */
	, EPCF_Not_Readable_File	/* Unreadable CSV file */
	, EPCF_Cannot_Open_File		/* Can't open CSV file */
	, EPCF_Cannot_Reallocate_Parsed_CSV_t_Lines_Vector
					/* Can't extend the allocated
					 * area for CSV-line vector
					 */
	, EPCF_Cannot_Allocate_Parsed_CSV_Line_t
					/* Can't allocate the area for
					 * CSV-line read-in buffer
					 */
	, EPCF_Broken_CSV_File		/* broken CSV file;
					 * found a line that isn't
					 * terminated by CR-LF sequence
					 */
	, EPCF_Cannot_Allocate_Memory 	/* Can't allocate several
					 * memories
					 */
	, EPCF_Failed_Convert_CSV_Line	/* Failed in converting the CSV
					 * line (internal fatal error)
					 */
	, EPCF_Zero_Size_File		/* the size of CSV-file is zero
					 */
} Errcode_Parse_CSV_File_t;

/* define prototypes */
extern char *string_to_csv_string(const char	*string
				, char		*csv
				, size_t	*csv_str_sz
				);

extern char *strings_to_csv_line(char		*strings[]
				, int		n
				, char		*csv_line
				, size_t	*csv_line_sz
				, char		delimit
					/* DLMT_COMMA or DLMT_TAB */
				);

extern char *get_csv_line(FILE		*stream
			, char		*csvl
			, size_t	*csvlsz
			, int		*is_broken
			);

extern char **csv_line_to_strings(char	*csvl
				, char	delimitch
				, int	*colscnt
				);

extern void free_parsed_csv_t(Parsed_CSV_t *parsed);

extern Errcode_Parse_CSV_File_t parse_csv_file(const char	*path
						, char		delimit
						, Parsed_CSV_t	*parsed
						);

extern char *get_value_parsed_csv_t(Parsed_CSV_t	*pcp
				, int			row
				, int			col
				);


/*
 * This API writes the CSV cell data, those are C-strings,
 * to `file_path' CSV file.
 *
 * Variable arguments describes the direction to write dart's.
 * o CSV_END means end of cells and directive to close the file.
 * o CSV_CONTINUE means the directives continues to the next invocation.
 *   (the following invocation has NULL at the 'file_path' argument)
 * o CSV_EOL means writing end of record ("\r\n") to the CSV file.
 * o CSV_EMPTY means writes several empty cells.
 *   This directive will be described as...
 * 	`CSV_EMPTY, <number of empty cells>,'
 *   (<number of empty cells> must be zero or positive integer.)
 * o CSV_COLUMN means writes a CSV cell.
 *   This directive will be described as...
 *      `CSV_COLUMN, <number of output characters>, <data of a cell>,'
 *   (<number of output characters> limits the output length.
 *    <data of a cell> points the string.)
 */
extern FILE *write_csv_columns(const char	*file_path
				, const char	delimit
				, FILE		*fp
				, long		*column_pos
				, ...
				);
#define CSV_CONTINUE	1 /* continue to the next invocation */
#define CSV_COLUMN	2 /* describes as:
			   * CSV_COLUMN, <output length>, <cell string>
			   */
#define CSV_EMPTY	3 /* describe as:
			   * CSV_EMPTY, <number of empty cells>
			   */
#define CSV_EOL		4 /* write end of record sequence ("\r\n") */
#define CSV_END		5 /* end of data and close the file */

#ifdef __cplusplus
}
#endif
#endif /* !defined(_CSV_CONVERTER_H_) } */
